@@ -2,6 +2,9 @@ import inspect
 import re
 from functools import wraps
 from uuid import uuid4
+
+import pandas
+
 from .six import string_types, integer_types
 
 from .fields import (FIELD_DATAFRAME, FIELD_TEXT, FIELD_NUMERIC, FIELD_NO_INPUT,
@@ -260,6 +263,7 @@ class DataframeType(BaseType):
     def __init__(self, data):
         self.value = self._assert_valid_value_and_cast(data["value"])
         self.column_prefix_map = data.get("column_prefix_map", {})
+        self.relationship_data = data.get("relationship_data", {})
 
     def _assert_valid_value_and_cast(self, value):
         if not hasattr(value, '__iter__'):
@@ -576,6 +580,18 @@ class DataframeType(BaseType):
         self.value[f"result_{uuid4()}"] = results
         return True in results
 
+    @type_operator(FIELD_DATAFRAME)
+    def empty_within_except_last_row(self, other_value: dict):
+        target = self.replace_prefix(other_value.get("target"))
+        comparator = other_value.get("comparator")
+        # group all targets by comparator
+        grouped_target = self.value.groupby(comparator)[target]
+        # validate all targets except the last one
+        results = grouped_target.apply(lambda x: x[:-1]).apply(pd.isnull)
+        # extract values with corresponding indexes from results
+        self.value[f"result_{uuid4()}"] = results.reset_index(level=0, drop=True)
+        return True in results.values
+
     @type_operator(FIELD_DATAFRAME)
     def non_empty(self, other_value: dict):
         target = self.replace_prefix(other_value.get("target"))
@@ -583,6 +599,18 @@ class DataframeType(BaseType):
         self.value[f"result_{uuid4()}"] = results
         return True in results
 
+    @type_operator(FIELD_DATAFRAME)
+    def non_empty_within_except_last_row(self, other_value: dict):
+        target = self.replace_prefix(other_value.get("target"))
+        comparator = other_value.get("comparator")
+        # group all targets by comparator
+        grouped_target = self.value.groupby(comparator)[target]
+        # validate all targets except the last one
+        results = ~grouped_target.apply(lambda x: x[:-1]).apply(pd.isnull)
+        # extract values with corresponding indexes from results
+        self.value[f"result_{uuid4()}"] = results.reset_index(level=0, drop=True)
+        return not(False in results.values)
+
     @type_operator(FIELD_DATAFRAME)
     def contains_all(self, other_value: dict):
         target = self.replace_prefix(other_value.get("target"))
@@ -712,6 +740,40 @@ class DataframeType(BaseType):
         self.value[f"result_{uuid4()}"] = results
         return True in results
 
+    @type_operator(FIELD_DATAFRAME)
+    def is_valid_reference(self, other_value):
+        target = self.replace_prefix(other_value.get("target"))
+        results = self.value[target].isin(self.relationship_data)
+        self.value[f"result_{uuid4()}"] = results
+        return not(False in results.values)
+
+    @type_operator(FIELD_DATAFRAME)
+    def is_not_valid_reference(self, other_value):
+        target = self.replace_prefix(other_value.get("target"))
+        results = ~self.value[target].isin(self.relationship_data)
+        self.value[f"result_{uuid4()}"] = results
+        return True in results.values
+
+    @type_operator(FIELD_DATAFRAME)
+    def is_valid_relationship(self, other_value):
+        target = self.replace_prefix(other_value.get("target"))
+        value_column = self.replace_prefix(other_value.get("comparator"))
+        results = self.value.apply(lambda row: self.detect_reference(row, value_column, target), axis=1)
+        self.value[f"result_{uuid4()}"] = results
+        return not(False in results.values)
+
+    @type_operator(FIELD_DATAFRAME)
+    def is_not_valid_relationship(self, other_value):
+        target = self.replace_prefix(other_value.get("target"))
+        value_column = self.replace_prefix(other_value.get("comparator"))
+        results = ~self.value.apply(lambda row: self.detect_reference(row, value_column, target), axis=1)
+        self.value[f"result_{uuid4()}"] = results
+        return True in results.values
+
+    def detect_reference(self, row, value_column, target_column):
+        target_data = self.relationship_data.get(row[target_column], pd.Series([]).values)
+        value = row[value_column]
+        return (value in target_data) or (value in target_data.astype(int).astype(str)) or (value in target_data.astype(str))
 
 @export_type
 class GenericType(SelectMultipleType, SelectType, StringType, NumericType, BooleanType):
